@@ -1,11 +1,3 @@
-#
-# Copyright (C) 2014 by Tomasz Konojacki
-#
-# This library is free software; you can redistribute it and/or modify
-# it under the same terms as Perl itself, either Perl version 5.18.2 or,
-# at your option, any later version of Perl 5 you may have available.
-#
-
 package Tie::ConfigFile;
 
 use strict;
@@ -14,15 +6,7 @@ use warnings;
 use IO::File;
 use Carp;
 
-our $VERSION = '0.02';
-
-sub __error {
-    my($self, $err) = @_;
-
-    croak $err if $self->{die_on_error};
-
-    return
-}
+our $VERSION = '1.00';
 
 sub __init {
     my $self = shift;
@@ -31,29 +15,70 @@ sub __init {
         $self->{filename},
         ($self->{readonly} ? O_RDONLY : O_RDWR) |
         ($self->{create_file} ? O_CREAT : 0)
-    ) or return __error($self, 'Unable to open file');
+    ) or croak sprintf('Unable to open file "%s".', $self->{filename});
 
     # we utf-8, baby
     $self->{_fh}->binmode(':utf8');
 
+    # init cache
+    $self->{_cache} = {};
+
     return 1;
 }
 
 sub __read {
     my $self = shift;
 
-    # __init will be launched only if there is no filehandle in $self->{_fh},
-    # if __init fails we return undef
-    return unless defined($self->{_fh}) || $self->__init;
+    # separator character
+    my $SEP = '=';
+    # comment character
+    my $COMM = ';';
 
-    my $s = $self->{separator};
+    my $cache = $self->{_cache};
 
+    my $i = 0;
     while (defined(my $line = $self->{_fh}->getline)) {
         chomp $line;
 
-        my($r_key, $r_value) = split /\s*$s\s*/, $line, 2;
+        # is it comment?
+        if ($line =~ /^\s*$COMM/) {
+            push @{ $cache->{structure} },
+                 {
+                    type => 'comment',
+                    value => $line,
+                 };
+        }
+        # is it just whitespace?
+        elsif ($line =~ /^\s*$/) {
+            push @{ $cache->{structure} },
+            {
+                type => 'whitespace',
+                value => $line,
+            }
+        }
+        # is it key-value pair?
+        elsif ($line =~ /$SEP/) {
+            my($key, $value) = split /\s*$SEP\s*/, $line, 2;
 
-        $self->{_cache}->{$r_key} = $r_value;
+            push @{ $cache->{structure} },
+            {
+                type => 'value',
+                value => $line,
+            };
+
+            # these are needed for fast retrieval of values
+            $cache->{values}{$key} = $value;
+            $cache->{key_to_struct}{$key} = $cache->{structure}->[$i];
+        }
+        # we can't parse it
+        else {
+            croak sprintf(
+                "Found garbage at <%s>, line %d.",
+                $self->{filename}, $i + 1
+            );
+        }
+
+        $i++;
     }
 
     return
@@ -62,13 +87,16 @@ sub __read {
 sub __write {
     my $self = shift;
 
+    my $cache = $self->{_cache};
+
     # clear the file before we write to it.
     $self->{_fh}->truncate(0);
     $self->{_fh}->seek(0, 0);
 
-    for (keys %{$self->{_cache}}) {
-        $self->{_fh}->print($_ , $self->{separator},
-                            $self->{_cache}->{$_}, "\n")
+    for my $item (@{ $cache->{structure} }) {
+        next if $item->{deleted};
+
+        $self->{_fh}->print($item->{value}, "\n");
     }
 
     # Due to flushing, writes are reliable even if user forgets to untie hash.
@@ -81,21 +109,19 @@ sub TIEHASH {
     my($class, %args) = @_;
 
     my $self = {
-        filename       => undef,
-        die_on_error   => 1,
-        readonly       => 1,
-        create_file    => 0,
-        empty_is_undef => 1,
-        separator      => '=',
+        filename         => undef,
+        readonly         => 1,
+        create_file      => 0,
         %args
     };
 
     unless (defined $self->{filename}) {
-        return __error($self, 'Filename is not specified!');
+        croak 'Filename is required.';
     }
 
     bless $self, $class;
 
+    $self->__init;
     $self->__read;
 
     return $self;
@@ -104,20 +130,38 @@ sub TIEHASH {
 sub FETCH {
     my($self, $key) = @_;
 
-    my $val = $self->{_cache}->{$key};
+    my $cache = $self->{_cache};
 
-    return if defined($val) && ($val eq '') && $self->{empty_is_undef};
-    return $val;
+    return $cache->{values}{$key};
 }
 
 sub STORE {
     my($self, $key, $value) = @_;
 
+    my $cache = $self->{_cache};
+
     if ($self->{readonly}) {
-        return $self->__error('STORE is not allowed on read-only config file')
+        croak 'STORE is not allowed on read-only config file.';
     }
 
-    $self->{_cache}->{$key} = $value;
+    if (exists $self->{_cache}{values}{$key}) {
+        # key already exist, just update it
+        $cache->{values}{$key} = $value;
+        $cache->{key_to_struct}{$key}{value} = "$key=$value",
+    }
+    else {
+        # create new key
+        push @{ $cache->{structure} },
+        {
+            type => 'value',
+            value => "$key=$value",
+        };
+        # index of element we just pushed
+        my $index = $#{ $cache->{structure} };
+
+        $cache->{values}{$key} = $value;
+        $cache->{key_to_struct}{$key} = $cache->{structure}->[$index],
+    }
 
     $self->__write;
 
@@ -127,11 +171,15 @@ sub STORE {
 sub DELETE {
     my($self, $key) = @_;
 
+    my $cache = $self->{_cache};
+
     if ($self->{readonly}) {
-        return $self->__error('DELETE is not allowed on read-only config file')
+        croak 'DELETE is not allowed on read-only config file.';
     }
 
-    delete $self->{_cache}->{$key};
+    $cache->{key_to_struct}{$key}{deleted} = 1;
+    delete $cache->{key_to_struct}{$key};
+    delete $cache->{values}{$key};
 
     $self->__write;
 
@@ -151,19 +199,19 @@ sub EXISTS {
     my($self, $key) = @_;
 
     # Empty keys have '' value, so if value is undefined, key doesn't exist.
-    return defined $self->{_cache}->{$key};
+    return exists $self->{_cache}{values}{$key};
 }
 
 sub FIRSTKEY {
     my $self = shift;
 
-    return (each %{$self->{_cache}})
+    return (each %{ $self->{_cache}{values} } )
 }
 
 sub NEXTKEY {
     my $self = shift;
 
-    return (each %{$self->{_cache}})
+    return (each %{ $self->{_cache}{values} } )
 }
 
 sub UNTIE {
@@ -182,7 +230,7 @@ sub DESTROY {
     return
 }
 
-'AKZ18/295-200P';
+'thank you based god';
 
 __END__
 
@@ -204,8 +252,18 @@ Tie::ConfigFile - Tie configuration file to a hash
 =head1 DESCRIPTION
 
 This module allows you to tie configuration file to a hash. To understand what
-"tie" means in this context, read L<perltie>. Comments in configuration files
-are B<NOT> supported.
+"tie" means in this context, read L<perltie>. Comments, empty lines and order
+in configuration file are always preserved. Formatting of a line is preserved
+unless you modify its value.
+
+=head1 EXAMPLE CONFIG FILE
+
+    key=value
+    ;comment
+    another_key = value
+    key with spaces=value with spaces
+
+    key after break=1
 
 =head1 OPTIONS
 
@@ -222,23 +280,8 @@ it doesn't exist.
 
 =item *
 
-C<die_on_error> (boolean, default: C<1>) - Croak when error happens.
-
-=item *
-
-C<empty_is_undef> (boolean, default: C<1>) - If key does exist but there is no
-value, return undef on retrieval.
-
-=item *
-
 C<readonly> (boolean, default: C<1>) - Disallow writing to the config file.
 
-=item *
-
-C<separator> (string, default: C<=>) - Key and value separator in config file.
-If specified separator and separator used in config file are different, said
-file will get corrupted on write.
-
 =back
 
 =head1 EXPORT
@@ -247,9 +290,17 @@ Nothing is exported.
 
 =head1 CAVEATS
 
+=over 4
+
+=item *
+
 When more than one process uses configuration file in non-readonly mode, data
 loss may happen.
 
+=item *
+
+Multidimensional hashes are not supported.
+
 =head1 SEE ALSO
 
 =over 4
@@ -276,7 +327,7 @@ DBD::IngresII Fossil repository is hosted at xenu.tk:
 
 =head1 COPYRIGHT AND LICENSE
 
-Copyright (C) 2014 by Tomasz Konojacki
+Copyright (C) 2015 by Tomasz Konojacki
 
 This library is free software; you can redistribute it and/or modify
 it under the same terms as Perl itself, either Perl version 5.18.2 or,
